@@ -1,5 +1,7 @@
 #include "stdafx.h"
+#include <sstream>
 #include <math.h>
+
 #include "alglibmisc.h"
 #include "alglibinternal.h"
 #include "linalg.h"
@@ -1219,6 +1221,318 @@ int main()
             return 1;
     }
         
+    
+    //
+    // Serialization properties
+    //
+    {
+        //
+        // Test kd-tree serialization
+        //
+        bool passed = true;
+        alglib::hqrndstate rs;
+        alglib::kdtree tree0;
+        alglib::real_2d_array xy, rxy0, rxy1;
+        alglib::real_1d_array qx;
+        const int npts = 50;
+        const int nx = 2;
+        const int ny = 1;
+        int cnt0, cnt1;
+        alglib::hqrndrandomize(rs);
+        xy.setlength(npts, nx+ny);
+        for(int i=0; i<npts; i++)
+            for(int j=0; j<nx+ny; j++)
+                xy[i][j] = alglib::hqrndnormal(rs);
+        alglib::kdtreebuild(xy, npts, nx, ny, 2, tree0);
+        qx.setlength(nx);
+        
+        try
+        {
+            // test string serialization/unserialization
+            alglib::kdtree tree1;
+            std::string s;
+            alglib::kdtreeserialize(tree0, s);
+            alglib::kdtreeunserialize(s, tree1);
+            for(int i=0; i<100; i++)
+            {
+                for(int j=0; j<nx; j++)
+                    qx[j] = alglib::hqrndnormal(rs);
+                cnt0 = alglib::kdtreequeryknn(tree0, qx, 1, true);
+                cnt1 = alglib::kdtreequeryknn(tree1, qx, 1, true);
+                if( (cnt0!=1) || (cnt1!=1) )
+                {
+                    passed = false;
+                    break;
+                }
+                alglib::kdtreequeryresultsxy(tree0, rxy0);
+                alglib::kdtreequeryresultsxy(tree1, rxy1);
+                for(int j=0; j<nx+ny; j++)
+                    passed = passed && (rxy0[0][j]==rxy1[0][j]);
+            }
+        }
+        catch(...)
+        { passed = false; }
+        
+        try
+        {
+            // test stream serialization/unserialization
+            //
+            // NOTE: we add a few symbols at the beginning and after the end of the data
+            //       in order to test algorithm ability to work in the middle of the stream
+            alglib::kdtree tree1;
+            std::stringstream s;
+            s.put('b');
+            s.put('e');
+            s.put('g');
+            alglib::kdtreeserialize(tree0, s);
+            s.put('e');
+            s.put('n');
+            s.put('d');
+            s.seekg(0);
+            passed = passed && (s.get()=='b');
+            passed = passed && (s.get()=='e');
+            passed = passed && (s.get()=='g');
+            alglib::kdtreeunserialize(s, tree1);
+            passed = passed && (s.get()=='e');
+            passed = passed && (s.get()=='n');
+            passed = passed && (s.get()=='d');
+            for(int i=0; i<100; i++)
+            {
+                for(int j=0; j<nx; j++)
+                    qx[j] = alglib::hqrndnormal(rs);
+                cnt0 = alglib::kdtreequeryknn(tree0, qx, 1, true);
+                cnt1 = alglib::kdtreequeryknn(tree1, qx, 1, true);
+                if( (cnt0!=1) || (cnt1!=1) )
+                {
+                    passed = false;
+                    break;
+                }
+                alglib::kdtreequeryresultsxy(tree0, rxy0);
+                alglib::kdtreequeryresultsxy(tree1, rxy1);
+                for(int j=0; j<nx+ny; j++)
+                    passed = passed && (rxy0[0][j]==rxy1[0][j]);
+            }
+        }
+        catch(...)
+        { passed = false; }
+        
+        try
+        {
+            // test string-to-stream serialization/unserialization
+            alglib::kdtree tree1;
+            std::string s0;
+            alglib::kdtreeserialize(tree0, s0);
+            std::stringstream s1(s0);
+            alglib::kdtreeunserialize(s1, tree1);
+            for(int i=0; i<100; i++)
+            {
+                for(int j=0; j<nx; j++)
+                    qx[j] = alglib::hqrndnormal(rs);
+                cnt0 = alglib::kdtreequeryknn(tree0, qx, 1, true);
+                cnt1 = alglib::kdtreequeryknn(tree1, qx, 1, true);
+                if( (cnt0!=1) || (cnt1!=1) )
+                {
+                    passed = false;
+                    break;
+                }
+                alglib::kdtreequeryresultsxy(tree0, rxy0);
+                alglib::kdtreequeryresultsxy(tree1, rxy1);
+                for(int j=0; j<nx+ny; j++)
+                    passed = passed && (rxy0[0][j]==rxy1[0][j]);
+            }
+        }
+        catch(...)
+        { passed = false; }
+        
+        try
+        {
+            // test stream-to-string serialization/unserialization
+            alglib::kdtree tree1;
+            std::stringstream s0;
+            alglib::kdtreeserialize(tree0, s0);
+            std::string s1 = s0.str();
+            alglib::kdtreeunserialize(s1, tree1);
+            for(int i=0; i<100; i++)
+            {
+                for(int j=0; j<nx; j++)
+                    qx[j] = alglib::hqrndnormal(rs);
+                cnt0 = alglib::kdtreequeryknn(tree0, qx, 1, true);
+                cnt1 = alglib::kdtreequeryknn(tree1, qx, 1, true);
+                if( (cnt0!=1) || (cnt1!=1) )
+                {
+                    passed = false;
+                    break;
+                }
+                alglib::kdtreequeryresultsxy(tree0, rxy0);
+                alglib::kdtreequeryresultsxy(tree1, rxy1);
+                for(int j=0; j<nx+ny; j++)
+                    passed = passed && (rxy0[0][j]==rxy1[0][j]);
+            }
+        }
+        catch(...)
+        { passed = false; }
+            
+        //
+        // Report
+        //
+        printf(fmt_str, "* Serialization (kd-tree)", passed ? "OK" : "FAILED");
+        fflush(stdout);
+        if( !passed )
+            return 1;
+    }
+    {
+        //
+        // Test legacy RBF interface
+        //
+        const char *pc_str = "50000000000 00000000000 20000000000 10000000000 A0000000000 \
+30000000000 20000000000 00000000000 A0000000000 30000000000\r\
+00000000000 20000000000 A0000000000 60000000000 00000000000\n\
+00000000000 00000000000 00000000000 00000000000 00000000000\r\n\
+00000000m_3 00000000000 00000000000 00000000m_3 00000000000\n\r\
+00000000000 00000000004 00000000000 00000000000\t00000000004 \
+00000000000 00000000000 00000000804 00000000000 00000000000 \
+00000000804 00000000000 00000000000 00000000G04 00000000000 \
+00000000000 00000000G04 00000000000 00000000000 00000000O04 \
+00000000000 00000000000 00000000O04 00000000000 00000000000 \
+00000000S04 00000000000 00000000000 00000000S04 00000000000 \
+00000000000 00000000W04 00000000000 00000000000 00000000W04 \
+00000000000 00000000000 00000000Y04 00000000000 00000000000 \
+00000000Y04 00000000000 00000000000 00000000K04 00000000000 \
+00000000000 00000000K04 00000000000 00000000000 A0000000000 \
+00000000000 10000000000 20000000000 30000000000 40000000000 \
+60000000000 70000000000 80000000000 90000000000 50000000000 \
+30000000000 00000000000 00000000000 00000000000 30000000000 \
+00000000Y04 00000000000 00000000000 u1000000000 00000000000 \
+00000000000 00000000000 60000000000 80000000000 00000000000 \
+50000000000 00000000000 50000000000 50000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 K0000000000 \
+00000000I04 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+00000000000 00000000000 00000000000 00000000000 00000000000 \
+A0000000000 30000000000 00000000000 00000000000 00000000000 \
+00000000m_3 00000000000 00000000000 00000000004 00000000000 \
+00000000000 00000000804 00000000000 00000000000 00000000G04 \
+00000000000 00000000000 00000000K04 00000000000 00000000000 \
+00000000O04 00000000000 00000000000 00000000S04 00000000000 \
+00000000000 00000000W04 00000000000 00000000000 00000000Y04 \
+00000000000 00000000000 A0000000000 40000000000 00000000q04 \
+-pAGQnQBI14 UqUWierJ91C esm8ag6G61C 00000000q04 4wcFMyCtu04 \
+oPDvwHqst04 CExQXp8Ct04 00000000q04 litzPFhRb0C oKJvjcct314 \
+5-fT-X8w614 00000000q04 3HSOsPVH11C vZWf4dgfv04 GbZg4MTJn04 \
+00000000q04 iv7rMhuR71C hRtixp15r_3 EvCEDtLu-0C 00000000q04 \
+41CXzA_q71C umRYLK2yp0C 1zzY3Zqd91C 00000000q04 JvxJzDeI21C \
+TVbyd7Ygz0C JLywRdR1n0C 00000000q04 KmFarhc4g0C 1ehrn2tUt0C \
+AECfwTIX814 00000000q04 Big__6hwt04 nSPzmAQrh_B 2H3o-KftH14 \
+00000000q04 n1b9361vI14 mhJhviUE114 54a_qyBrH1C 00000000q04 \
+10000000000 40000000000 StLCgor39-3 00000000000 00000000000 \
+6qTG7Ae-1_3\n";
+        alglib::real_1d_array ref_val("[-0.042560546916643, 0.942523544654062, 0.875197036560778, 0.0656948997826632, -0.743065973803404, -0.8903682039297, -0.26994815318748, 0.602248517290195, 0.980011992233124, 0.436594293214176]");
+        bool passed = true;
+        
+        try
+        {
+            // test unserialization from string without trailing end-of-stream symbol (dot)
+            // this test is necessary for backward compatibility
+            double eps = 0.0000000001;
+            alglib::rbfmodel model;
+            alglib::rbfunserialize(std::string(pc_str), model);
+            for(int i=0; i<ref_val.length(); i++)
+                passed = passed && (fabs(alglib::rbfcalc2(model,i,0)-ref_val[i])<eps);
+        }
+        catch(...)
+        { passed = false; }
+        
+        try
+        {
+            // test unserialization from string with trailing end-of-stream symbol (dot)
+            // this test is necessary for forward compatibility
+            double eps = 0.0000000001;
+            alglib::rbfmodel model;
+            alglib::rbfunserialize(std::string(pc_str)+".", model);
+            for(int i=0; i<ref_val.length(); i++)
+                passed = passed && (fabs(alglib::rbfcalc2(model,i,0)-ref_val[i])<eps);
+        }
+        catch(...)
+        { passed = false; }
+        
+        try
+        {
+            // test unserialization from stream WITHOUT trailing end-of-stream symbol (dot)
+            // this test MUST fail
+            double eps = 0.0000000001;
+            std::string _s(pc_str);
+            std::istringstream stream(_s);
+            alglib::rbfmodel model;
+            alglib::rbfunserialize(stream, model);
+            passed = false;
+        }
+        catch(...)
+        { /* do nothing, it is expected to fail */ }
+        
+        try
+        {
+            // test unserialization from stream WITH trailing end-of-stream symbol (dot)
+            // this test must succeed
+            double eps = 0.0000000001;
+            std::string _s = std::string(pc_str)+".";
+            std::istringstream stream(_s);
+            alglib::rbfmodel model;
+            alglib::rbfunserialize(stream, model);
+            for(int i=0; i<ref_val.length(); i++)
+                passed = passed && (fabs(alglib::rbfcalc2(model,i,0)-ref_val[i])<eps);
+        }
+        catch(...)
+        { passed = false; }
+        
+        try
+        {
+            // test that we can read from the stream after unserialization
+            double eps = 0.0000000001;
+            std::string _s = std::string(pc_str)+".<az>";
+            std::istringstream stream(_s);
+            alglib::rbfmodel model;
+            alglib::rbfunserialize(stream, model);
+            for(int i=0; i<ref_val.length(); i++)
+                passed = passed && (fabs(alglib::rbfcalc2(model,i,0)-ref_val[i])<eps);
+            passed = passed && (stream.get()=='<');
+            passed = passed && (stream.get()=='a');
+            passed = passed && (stream.get()=='z');
+            passed = passed && (stream.get()=='>');
+        }
+        catch(...)
+        { passed = false; }
+            
+        //
+        // Report
+        //
+        printf(fmt_str, "* Serialization (RBF)", passed ? "OK" : "FAILED");
+        fflush(stdout);
+        if( !passed )
+            return 1;
+    }
+
     //
     // Testing issues which must be fixed
     //
@@ -1232,18 +1546,18 @@ int main()
         {
             //
             // Test CG
-            // Stopping criteria - after |grad|<epsG
+            // Stopping criteria - EpsX
             //
             mincgstate state;
             mincgreport rep;
             real_1d_array x = "[0.0]";
             double x0 = 20*alglib::randomreal()-10;
-            double epsg = 1.0E-9;
+            double epsx = 1.0E-9;
             mincgcreate(1, x, state);
-            mincgsetcond(state, epsg, 0.0, 0.0, 0);
+            mincgsetcond(state, 0.0, 0.0, epsx, 0);
             mincgoptimize(state, func505_grad, NULL, &x0);
             mincgresults(state, x, rep);
-            issue505_passed = issue505_passed && (fabs(4*pow(x[0]-x0,3))<epsg);
+            issue505_passed = issue505_passed && (fabs(4*pow(x[0]-x0,3))<1.0E-3);
         }
         catch(...)
         { issue505_passed = false; }
@@ -1257,9 +1571,9 @@ int main()
             minlmreport rep;
             real_1d_array x = "[0.0]";
             double x0 = 20*alglib::randomreal()-10;
-            double epsg = 1.0E-9;
+            double epsx = 1.0E-9;
             minlmcreatevj(1, 2, x, state);
-            minlmsetcond(state, epsg, 0.0, 0.0, 0);
+            minlmsetcond(state, epsx, 0);
             minlmoptimize(state, func505_vec, func505_jac, NULL, &x0);
             minlmresults(state, x, rep);
             issue505_passed = issue505_passed && (fabs(x[0]-x0)<1.0E-3);
